/*
******************************************************************************
*
*      File             : oacs.c
*      Version          : 2.0.0
*      Date             : 27 November 2000
*      Author           : Jochen Junkawitsch
*      Description      : source file for OACS selection rules
*
*
******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include "oacs.h"


/* ---------------------------------------------------------------------------------------
                                 private functions
   --------------------------------------------------------------------------------------- */

void      selectMode     (CodecMode codecMode, CodecSet* cscs, CodecSet* oacs);
void      notInclude     (CodecMode codecMode, CodecSet* cscs);
int       ready          (CodecSet cscs, CodecSet oacs, int cmacs);
CodecMode findHighest    (CodecSet cs);
CodecMode findLowest     (CodecSet cs);
void      selectLowest   (CodecSet* cscs, CodecSet* oacs);
void      selectHighest  (CodecSet* cscs, CodecSet* oacs);
CodecSet  calc_OACS_FR   (CodecSet cscs, int cmacs);
CodecSet  calc_OACS_HR   (CodecSet cscs, int cmacs);
void      checkDefaultFR (CodecSet* cscs, CodecSet* oacs, int cmacs);
void      checkDefaultHR (CodecSet* cscs, CodecSet* oacs, int cmacs);

/* ---------------------------------------------------------------------------------------

   --------------------------------------------------------------------------------------- */


CodecSet calc_OACS(CodecSet cscs, int cmacs, CodecType codecType)
{
  switch (codecType)
  {
    case FR_AMR: return calc_OACS_FR(cscs, cmacs);
    case HR_AMR: return calc_OACS_HR(cscs, cmacs);
    default: fprintf(stderr, "ERROR: invalid codec type.\n"); exit(0);
  }
  return (CodecType) 0;
}


CodecSet calc_OACS_FR(CodecSet cscs, int cmacs)
{
  CodecSet oacs = 0;
    
  while (1)
  {
	  
  
	  /*
		if CMACS == 1
		select mode according to preference list {6.7, 7.4, 5.9, 5.15, 4.75, 7.95, 10.2, 12.2}
		*/
	  if (cmacs == 1)
	  {
		  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_102, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_122, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  fprintf(stderr, "ERROR: This point must never be reached.\n"); exit(0);
	  }
	  
	  
	  /*
		if CMACS == 2
		*/
	  if (cmacs == 2)
	  {
		  /* If mode 10.2 is supported, do not include mode 12.2. */
		  if (cscs & mode_102) notInclude(mode_122, &cscs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* Select highest mode. */
		  selectHighest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* If mode 12.2 or mode 10.2 is selected, select mode according to
			 preference list {6.7, 7.4, 5.9, 5.15, 4.75, 7.95, 10.2, 12.2}. */
		  if (oacs & (mode_122 | mode_102))
		  {
			  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_102, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_122, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  }
		  
		  /* Select lowest mode. */
		  selectLowest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  fprintf(stderr, "ERROR: This point must never be reached.\n"); exit(0);
	  }
	  
	  
	  /* CMACS > 2 */
	  
	  /* if 10.2 is supported, do not include 12.2 */
	  if (cscs & mode_102) notInclude(mode_122, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if 4.75 is supported, do not include 5.15 */
	  if (cscs & mode_475) notInclude(mode_515, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.15 is supported, do not include mode 5.9. */
	  if (cscs & mode_515) notInclude(mode_590, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.9 is supported and mode 4.75 is not supported, do not include mode 6.7. */
	  if ((cscs & mode_590) && !(cscs & mode_475)) notInclude(mode_670, &cscs); 
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if (12.2 or 10.2) and 7.4 is supported, do not include 7.95 */
	  if ((cscs & (mode_122 | mode_102)) && (cscs & mode_740)) notInclude(mode_795, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 7.95 is supported, do not include 7.4. */
	  if (cscs & mode_795) notInclude(mode_740, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select lowest mode */
	  selectLowest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select highest mode */
	  selectHighest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select 6.7 */
	  selectMode(mode_670, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select 5.9 */
	  selectMode(mode_590, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  
	  /* ----- */
	  
	  fprintf(stderr, "ERROR: This point must never be reached.\n");

  } /* end of while (1) */
  
  if (nModes(oacs) < cmacs)
	  oacs |= cscs;

  return oacs;
}


CodecSet calc_OACS_HR(CodecSet cscs, int cmacs)
{
  CodecSet oacs = 0;
    
  while (1)
  {
	  
  
	  /*
		if CMACS == 1
		select mode according to preference list {5.9, 5.15, 4.75, 6.7, 7.4, 7.95}
		*/
	  if (cmacs == 1)
	  {
		  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  fprintf(stderr, "ERROR: This point must never be reached.\n"); exit(0);
	  }
	  
	  
	  /*
		if CMACS == 2
		*/
	  if (cmacs == 2)
	  {
		  /* Select highest mode. */
		  selectHighest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* Select lowest mode. */
		  selectLowest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  fprintf(stderr, "ERROR: This point must never be reached.\n"); exit(0);
	  }
	  
	  
	  /* CMACS > 2 */
	  
	  /* if 10.2 is supported, do not include 12.2 */
	  if (cscs & mode_102) notInclude(mode_122, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if 4.75 is supported, do not include 5.15 */
	  if (cscs & mode_475) notInclude(mode_515, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.15 is supported, do not include mode 5.9. */
	  if (cscs & mode_515) notInclude(mode_590, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.9 is supported and mode 4.75 is not supported, do not include mode 6.7. */
	  if ((cscs & mode_590) && !(cscs & mode_475)) notInclude(mode_670, &cscs); 
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if (12.2 or 10.2) and 7.4 is supported, do not include 7.95 */
	  if ((cscs & (mode_122 | mode_102)) && (cscs & mode_740)) notInclude(mode_795, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 7.95 is supported, do not include 7.4. */
	  if (cscs & mode_795) notInclude(mode_740, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select lowest mode */
	  selectLowest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select highest mode */
	  selectHighest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select 6.7 */
	  selectMode(mode_670, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select 5.9 */
	  selectMode(mode_590, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  
	  /* ----- */
	  
	  fprintf(stderr, "ERROR: This point must never be reached.\n");

  } /* end of while (1) */
  
  if (nModes(oacs) < cmacs)
	  oacs |= cscs;

  return oacs;
}


int nModes(CodecSet codecSet)
{
  int k, n=0;
  for (k=0; k<8; k++)
  {
	  if (codecSet & 1) { n++; }
	  codecSet >>= 1;
  }
  return n;
}

void selectMode(CodecMode codecMode, CodecSet* cscs, CodecSet* oacs)
{
	if (*cscs & codecMode)
	{
		*cscs &= codecMode ^ 0xFF;
		*oacs |= codecMode;
	}
}

void notInclude(CodecMode codecMode, CodecSet* cscs)
{
  *cscs &= codecMode ^ 0xFF;
}

int ready(CodecSet cscs, CodecSet oacs, int cmacs)
{
    if (nModes(oacs) == cmacs) return 1;
    if (nModes(oacs) + nModes(cscs) <= cmacs) return 1;
    return 0;
}


CodecMode findHighest(CodecSet cs)
{
  if (cs & mode_122) return mode_122;
  if (cs & mode_102) return mode_102;
  if (cs & mode_795) return mode_795;
  if (cs & mode_740) return mode_740;
  if (cs & mode_670) return mode_670;
  if (cs & mode_590) return mode_590;
  if (cs & mode_515) return mode_515;
  if (cs & mode_475) return mode_475;
  fprintf(stderr, "ERROR: This point must never be reached.\n"); exit(0);
  return (CodecMode) 0;
}

CodecMode findLowest(CodecSet cs)
{
  if (cs & mode_475) return mode_475;
  if (cs & mode_515) return mode_515;
  if (cs & mode_590) return mode_590;
  if (cs & mode_670) return mode_670;
  if (cs & mode_740) return mode_740;
  if (cs & mode_795) return mode_795;
  if (cs & mode_102) return mode_102;
  if (cs & mode_122) return mode_122;
  fprintf(stderr, "ERROR: This point must never be reached.\n"); exit(0);
  return (CodecMode) 0;
}



void selectLowest(CodecSet* cscs, CodecSet* oacs)
{
  selectMode(findLowest(*cscs), cscs, oacs);
}

void selectHighest(CodecSet* cscs, CodecSet* oacs)
{
  selectMode(findHighest(*cscs), cscs, oacs);
}

void checkDefaultFR(CodecSet* cscs, CodecSet* oacs, int cmacs)
{
  if ((cmacs >= 4) &&
      (*cscs & mode_475) &&
      (*cscs & mode_590) &&
      (*cscs & mode_670) &&
      (*cscs & mode_102))
  {
    selectMode(mode_475, cscs, oacs);
    selectMode(mode_590, cscs, oacs);
    selectMode(mode_670, cscs, oacs);
    selectMode(mode_102, cscs, oacs);
  }
}

void checkDefaultHR(CodecSet* cscs, CodecSet* oacs, int cmacs)
{
  if ((cmacs >= 4) &&
      (*cscs & mode_475) &&
      (*cscs & mode_590) &&
      (*cscs & mode_670) &&
      (*cscs & mode_740))
  {
    selectMode(mode_475, cscs, oacs);
    selectMode(mode_590, cscs, oacs);
    selectMode(mode_670, cscs, oacs);
    selectMode(mode_740, cscs, oacs);
  }
}


int highModeRule(CodecSet lacs, CodecSet dacs, CodecSet oacs)
{
  CodecMode highOACS = findHighest(oacs);
  CodecMode highLACS = findHighest(lacs);
  CodecMode highDACS = findHighest(dacs);
  CodecMode minACSMode = (highLACS < highDACS ? highLACS : highDACS);
  
  return (highOACS >= (minACSMode >> 1));
}

int lowModeRule(CodecSet lacs, CodecSet dacs, CodecSet oacs, CodecType codecType)
{
  CodecMode max = (codecType == HR_AMR ? mode_590 : mode_740);
  CodecMode lowOACS = findLowest(oacs);
  CodecMode lowLACS = findLowest(lacs);
  CodecMode lowDACS = findLowest(dacs);
  CodecMode lowestACSMode = (lowLACS < lowDACS ? lowLACS : lowDACS);
  
  return (lowOACS <= max || lowOACS <= lowestACSMode);
}

